"""The prepare_output module."""

# --------------------------- Import libraries and functions --------------------------
from typing import Optional

from functools import partial

import pandas as pd

from pycvcqv.formulas import _cqv, _cv
from pycvcqv.types import PoolTypeT


# -------------------------------- function definition --------------------------------
def prepare_cv_datafame(
    pool: PoolTypeT,
    data: pd.DataFrame,
    ddof: Optional[int] = 1,
    skipna: Optional[bool] = True,
    ndigits: Optional[int] = 4,
    correction: Optional[bool] = False,
    multiplier: Optional[int] = 1,
) -> pd.DataFrame:
    """Prepares result pd.DataFrame for cv."""
    data = data.select_dtypes(include="number")
    result = pd.DataFrame(
        {
            "columns": data.columns,
            "cv": pool.map(
                partial(
                    _cv,
                    ddof=ddof,
                    skipna=skipna,
                    ndigits=ndigits,
                    correction=correction,
                    multiplier=multiplier,
                ),
                (pd.Series(data.loc[:, col]) for col in data.columns),
            ),
        }
    )
    return result


def prepare_cqv_datafame(
    pool: PoolTypeT,
    data: pd.DataFrame,
    ndigits: Optional[int] = 4,
    interpolation: Optional[str] = "linear",
    multiplier: Optional[int] = 1,
) -> pd.DataFrame:
    """Prepares result pd.DataFrame for cqv."""
    data = data.select_dtypes(include="number")
    result = pd.DataFrame(
        {
            "columns": data.columns,
            "cqv": pool.map(
                partial(
                    _cqv,
                    ndigits=ndigits,
                    interpolation=interpolation,
                    multiplier=multiplier,
                ),
                (pd.Series(data.loc[:, col]) for col in data.columns),
            ),
        }
    )
    return result



"""Coefficient of Variation (CV) and Coefficient of Quartile Variation (CQV)."""

from typing import Any

import sys

from pycvcqv.cqv import cqv
from pycvcqv.cv import coefficient_of_variation
from pycvcqv.dataframe import cqv_dataframe, cv_dataframe
from pycvcqv.is_numeric import is_numeric

if sys.version_info >= (3, 8):  # pragma: no cover
    from importlib import metadata as importlib_metadata
else:  # pragma: no cover
    import importlib_metadata


# -------------------------------- function definition --------------------------------
def get_version() -> Any:
    """Gets the version of package."""
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()

__all__ = [
    "coefficient_of_variation",
    "is_numeric",
    "cqv",
    "cv_dataframe",
    "cqv_dataframe",
]



"""The dataframe module."""

# --------------------------- Import libraries and functions --------------------------
from typing import Optional

import pandas as pd

from pycvcqv.multithread import multithread_cqv_processor, multithread_cv_processor
from pycvcqv.singlethread import singlethread_cqv_processor, singlethread_cv_processor
from pycvcqv.userthread import userthread_cqv_processor, userthread_cv_processor


# -------------------------------- function definition --------------------------------
def cv_dataframe(
    data: pd.DataFrame,
    num_threads: Optional[int] = 1,
    ddof: Optional[int] = 1,
    skipna: Optional[bool] = True,
    ndigits: Optional[int] = 4,
    correction: Optional[bool] = False,
    multiplier: Optional[int] = 1,
) -> pd.DataFrame:
    """Selects method to perform cv calculations on dataframe or array-like objects."""
    operators = list(
        k
        for k, v in {
            multithread_cv_processor: -1,
            singlethread_cv_processor: 0,
            userthread_cv_processor: num_threads,
        }.items()
        if v == num_threads
    )
    return operators[0](
        data=data,
        num_threads=num_threads,
        ddof=ddof,
        skipna=skipna,
        ndigits=ndigits,
        correction=correction,
        multiplier=multiplier,
    )


def cqv_dataframe(
    data: pd.DataFrame,
    ndigits: Optional[int] = 4,
    interpolation: Optional[str] = "linear",
    multiplier: Optional[int] = 1,
    num_threads: Optional[int] = 1,
) -> pd.DataFrame:
    """Selects method to perform cqv calculations on dataframe or array-like objects."""
    # ------------------------------------ threads  -----------------------------------
    operators = list(
        k
        for k, v in {
            multithread_cqv_processor: -1,
            singlethread_cqv_processor: 0,
            userthread_cqv_processor: num_threads,
        }.items()
        if v == num_threads
    )
    return operators[0](data, ndigits, interpolation, multiplier, num_threads)



"""The types used in the package."""

# --------------------------- Import libraries and functions --------------------------
from typing import List, Tuple, TypeVar, Union

from multiprocessing import pool

import numpy as np
import pandas as pd
from numpy import typing as npt

# ---------------------------------- types definition ---------------------------------
ListFloat = List[float]
ListInt = List[int]
TupleFloat = Tuple[float]
TupleInt = Tuple[int]
ArrayFloat = npt.NDArray[np.float64]
ArrayInt = npt.NDArray[np.int_]
PoolTypeT = TypeVar("PoolTypeT", bound=pool.Pool)
NumArrayLike = TypeVar(
    "NumArrayLike",
    bound=Union[
        pd.Series, ArrayInt, ArrayFloat, ListFloat, ListInt, TupleFloat, TupleInt
    ],
)



"""Formulas of cvcqv."""

# --------------------------- Import libraries and functions --------------------------
from typing import Optional  # Optional type for function arguments.

import numpy as np  # To handle numeric infinity values.
import pandas as pd  # Data analysis and manipulation library.

from pycvcqv.types import NumArrayLike  # custom numeric array defined in types.py.


# -------------------------------- function definition --------------------------------
def _cv(
    data: NumArrayLike,
    ddof: Optional[int] = 1,
    skipna: Optional[bool] = True,
    ndigits: Optional[int] = 4,
    correction: Optional[bool] = False,
    multiplier: Optional[int] = 1,
) -> float:
    """Internal function to calculate cv."""
    _data: pd.Series = pd.Series(data)
    if _data.mean(skipna=skipna) == 0 or (  # check if the mean is 0
        _data.mean(skipna=skipna) < 0.000001  # check if the mean is close to 0
        # ----------------- also, check if the std is higher than mean ----------------
        and _data.std(skipna=skipna, ddof=ddof) > _data.mean(skipna=skipna)
    ):  # ------- The Inf value which comes from numpy to handle infinity cases -------
        return float(np.inf)  # to ensure that the returned infinity value is 'float'
    # ------------------ the basic coefficient of variation function ------------------
    _cv = _data.std(skipna=skipna, ddof=ddof) / _data.mean(skipna=skipna)
    length = len(_data)  # ---------- calculate the length of input array ---------
    if correction:  # -------------------- return the corrected cv --------------------
        corrected_rounded_cv: float = round(  # ----------- round the result ----------
            multiplier  # ----------- multiply the cv e.g, 100 for percentage ---------
            * (
                _cv
                * (  # --- this is the coefficient for the correction of sample size --
                    1  # ----------- the coefficient should be less than zero ----------
                    - ((4 * (length - 1)) ** (-1))  # ------------ 1/(4n-1) -----------
                    + ((_cv**2) * (length ** (-1)))  # ----------- (cv^2/n) -----------
                    + (2 * ((length - 1) ** (2))) ** (-1)  # ----- 1/2*(n-1)^2 --------
                )
            ),
            ndigits=ndigits,  # --------------- decimals for the round ----------------
        )
        return corrected_rounded_cv
    rounded_cv: float = round(  # ---------------- return the basic cv ----------------
        multiplier * (_cv),  # -------- multiply the cv e.g, 100 for percentage -------
        ndigits=ndigits,  # ------------------ decimals for the round -----------------
    )
    return rounded_cv


def _cqv(
    data: NumArrayLike,
    ndigits: Optional[int] = 4,
    interpolation: Optional[str] = "linear",
    multiplier: Optional[int] = 1,
) -> float:
    """Internal function to calculate cqv."""
    _data: pd.Series = pd.Series(data)
    # ---------------------- calculate the quantiles of the data ----------------------
    quantile1 = _data.quantile(0.25, interpolation=interpolation)  # q1 = p25
    quantile3 = _data.quantile(0.75, interpolation=interpolation)  # q3 = p75
    # ------------------- raise warning for 0 divisor when q3+q1 = 0 ------------------
    if quantile1 + quantile3 == 0:
        raise Warning("cqv is NaN because q3 and q1 are 0")
    _cqv = (quantile3 - quantile1) / (quantile3 + quantile1)  # ---- the basic cqv ----
    rounded_cqv: float = round(  # ------------------ return the cqv ------------------
        multiplier * _cqv,  # -------- multiply the cqv e.g, 100 for percentage -------
        ndigits=ndigits,  # ------------------ decimals for the round -----------------
    )
    return rounded_cqv



"""Coefficient of Variation (cv)."""

# --------------------------- Import libraries and functions --------------------------
from typing import Optional, Union

import pandas as pd

from pycvcqv.dataframe import cv_dataframe
from pycvcqv.formulas import _cv
from pycvcqv.is_numeric import is_numeric
from pycvcqv.types import ArrayFloat, ArrayInt, ListFloat, ListInt, TupleFloat, TupleInt

# -------------------------------- function definition --------------------------------


@is_numeric  # decorator to check whether the input vector is numeric
def coefficient_of_variation(
    data: Union[
        pd.Series,
        ArrayInt,
        ArrayFloat,
        ListFloat,
        ListInt,
        TupleFloat,
        TupleInt,
        pd.DataFrame,
    ],
    ddof: Optional[int] = 1,
    skipna: Optional[bool] = True,
    ndigits: Optional[int] = 4,
    correction: Optional[bool] = False,
    multiplier: Optional[int] = 1,
    num_threads: Optional[int] = 1,
) -> Union[float, pd.DataFrame]:
    """Coefficient of variation.

    Args:
        data (pandas.core.series.Series, numpy.ndarray, list, tuple, or pd.DataFrame,
            default numpy.ndarray): Having either float or integer elements. In
            dataframes, columns with numeric data will be used.
        ddof (int, default 1): Delta Degrees of Freedom, The divisor used in
            calculations is ``N - ddof``, where ``N`` represents the number of
            elements of the data. By default `ddof` is 1.
        skipna (bool, default True): Exclude NA/null values when computing
            the result.
        ndigits (int, default 4): Indicates the number of decimal places, from
            built-in function round in module builtins.
        correction (bool, default False): To account for the sample size,
            correction might be set to True.
        multiplier (int, default 1): cv will be multiplied by it, such as 100,
            when you want to report cv as percentage.
        num_threads (int, default 1): The number of therads to use. This speeds up
            calculation for the pd.DataFrame inputs. Defaults to single thread. If -1
            is specified then multiprocessing.cpu_count() is used instead.


    Returns:
        Union[float, pd.DataFrame]: the coefficient(s) of variation i.e.,
            sd(x)/mean(x).

    Examples:
        .. code:: python

            >>> coefficient_of_variation(
            ...     data=pd.Series([
            ...         0.2, 0.5, 1.1, 1.4, 1.8, 2.3, 2.5, 2.7, 3.5, 4.4,
            ...         4.6, 5.4, 5.4, 5.7, 5.8, 5.9, 6.0, 6.6, 7.1, 7.9
            ...     ]),
            ...     ddof=1,
            ...     correction=False,
            ...     multiplier=100
            ... )
            57.77
    """
    # ----------------------------------- DataFrame  ----------------------------------
    if isinstance(data, pd.DataFrame):
        result = cv_dataframe(
            data=data,
            num_threads=num_threads,
            ddof=ddof,
            skipna=skipna,
            ndigits=ndigits,
            correction=correction,
            multiplier=multiplier,
        )
    # --------------------------------- non DataFrame  --------------------------------
    else:
        result = float(
            _cv(
                data=data,
                ddof=ddof,
                skipna=skipna,
                ndigits=ndigits,
                correction=correction,
                multiplier=multiplier,
            )
        )
    return result



"""The singlethread module."""

# --------------------------- Import libraries and functions --------------------------
from typing import Optional

import multiprocessing as mp

import pandas as pd

from pycvcqv.prepare_output import prepare_cqv_datafame, prepare_cv_datafame


# -------------------------------- function definition --------------------------------
def singlethread_cv_processor(
    data: pd.DataFrame,
    num_threads: Optional[int] = 1,
    ddof: Optional[int] = 1,
    skipna: Optional[bool] = True,
    ndigits: Optional[int] = 4,
    correction: Optional[bool] = False,
    multiplier: Optional[int] = 1,
) -> pd.DataFrame:
    """Performs single thread cv for pd.DataFrame."""
    print(num_threads)
    with mp.Pool(1) as pool:
        result = prepare_cv_datafame(
            data=data,
            ddof=ddof,
            skipna=skipna,
            ndigits=ndigits,
            correction=correction,
            multiplier=multiplier,
            pool=pool,
        )
    pool.close()
    return result


def singlethread_cqv_processor(
    data: pd.DataFrame,
    ndigits: Optional[int] = 4,
    interpolation: Optional[str] = "linear",
    multiplier: Optional[int] = 1,
    num_threads: Optional[int] = 1,
) -> pd.DataFrame:
    """Performs single thread cqv for pd.DataFrame."""
    print(num_threads)
    with mp.Pool(1) as pool:
        result = prepare_cqv_datafame(
            data=data,
            ndigits=ndigits,
            interpolation=interpolation,
            multiplier=multiplier,
            pool=pool,
        )
    pool.close()
    return result



"""Coefficient of Quartile Variation (cqv)."""

# --------------------------- Import libraries and functions --------------------------
from typing import Optional, Union

import pandas as pd

from pycvcqv.dataframe import cqv_dataframe
from pycvcqv.formulas import _cqv
from pycvcqv.is_numeric import is_numeric
from pycvcqv.types import ArrayFloat, ArrayInt, ListFloat, ListInt, TupleFloat, TupleInt


# -------------------------------- function definition --------------------------------
@is_numeric  # decorator to check whether the input vector is numeric
def cqv(
    data: Union[
        pd.Series,
        ArrayFloat,
        ArrayInt,
        ListFloat,
        ListInt,
        TupleFloat,
        TupleInt,
        pd.DataFrame,
    ],
    ndigits: Optional[int] = 4,
    interpolation: Optional[str] = "linear",
    multiplier: Optional[int] = 1,
    num_threads: Optional[int] = 1,
) -> Union[pd.DataFrame, float]:
    """Coefficient of quartile variation.

    Args:
        data (pandas.core.series.Series, numpy.ndarray, list, tuple, or pd.DataFrame,
            default numpy.ndarray): Having either float or integer elements. In
            dataframes, columns with numeric data will be used.
        ndigits (int, default 4): Indicates the number of decimal places, from
            built-in function round in module builtins.
        interpolation (str, default 'linear'): It specifies the interpolation method to
            use, when the desired quantile lies between two data points `i` and `j`:
            * linear: `i + (j - i) * fraction`, where `fraction` is the
              fractional part of the index surrounded by `i` and `j`.
            * lower: `i`.
            * higher: `j`.
            * nearest: `i` or `j` whichever is nearest.
            * midpoint: (`i` + `j`) / 2.
        multiplier (int, default 1): cqv will be multiplied by it, such as 100,
            when you want to report cqv as percentage.
        num_threads (int, default 1): The number of therads to use. This speeds up
            calculation for the pd.DataFrame inputs. Defaults to single thread. If -1
            is specified then multiprocessing.cpu_count() is used instead.

    Returns:
        Union[pd.DataFrame, float]: the coefficient(s) of quartile variation.

    Examples:
        .. code:: python

            >>> cqv(
            ...     data=pd.Series([
            ...         0.2, 0.5, 1.1, 1.4, 1.8, 2.3, 2.5, 2.7, 3.5, 4.4,
            ...         4.6, 5.4, 5.4, 5.7, 5.8, 5.9, 6.0, 6.6, 7.1, 7.9
            ...     ]),
            ...     multiplier=100
            ... )
            45.625
    """
    # ----------------------------------- DataFrame  ----------------------------------
    if isinstance(data, pd.DataFrame):
        result = cqv_dataframe(
            data=data,
            num_threads=num_threads,
            ndigits=ndigits,
            interpolation=interpolation,
            multiplier=multiplier,
        )
    # --------------------------------- non DataFrame  --------------------------------
    else:
        result = float(_cqv(data, ndigits, interpolation, multiplier))
    return result
